@@ -9,6 +9,7 @@ import os
 
 import gi
 gi.require_version('AppStream', '1.0')
+gi.require_version('Gtk', '3.0')
 from gi.repository import AppStream, GLib, GObject, Gtk, Gio
 
 try:
@@ -23,15 +24,37 @@ from .dialogs import ChangesConfirmDialog, FlatpakProgressWindow
 from .misc import debug
 
 class FlatpakRemoteInfo():
-    def __init__(self, remote):
-        self.name = remote.get_name()
-        self.title = remote.get_title()
-        self.url = remote.get_url()
-        self.disabled = remote.get_disabled()
-        self.noenumerate = remote.get_noenumerate()
+    def __init__(self, remote=None):
+        if remote:
+            self.name = remote.get_name()
+            self.title = remote.get_title()
+            self.url = remote.get_url()
+            self.disabled = remote.get_disabled()
+            self.noenumerate = remote.get_noenumerate()
 
-        if not self.title or self.title == "":
-            self.title = self.name.capitalize()
+            if not self.title or self.title == "":
+                self.title = self.name.capitalize()
+        else:
+            self.name = None
+            self.title = None
+            self.url = None
+            self.disabled = False
+            self.noenumerate = False
+
+    @classmethod
+    def from_json(cls, json_data:dict):
+        inst = cls()
+
+        inst.name = json_data["name"]
+        inst.title = json_data["title"]
+        inst.url = json_data["url"]
+        inst.disabled = json_data["disabled"]
+        inst.noenumerate = json_data["noenumerate"]
+
+        return inst
+
+    def to_json(self):
+        return self.__dict__
 
 _fp_sys = None
 
@@ -68,7 +91,11 @@ def _get_remote_name_by_url(fp_sys, url):
         remotes = []
 
     for remote in remotes:
-        if remote.get_url() == url:
+        remote_url = remote.get_url()
+        if remote_url.endswith('/'): #flatpakrefs are often missing the trailing forward slash in the url
+            remote_url = remote_url[:-1]
+
+        if remote_url == url:
             name = remote.get_name()
 
     return name
@@ -80,7 +107,7 @@ def _get_file_timestamp(gfile):
         return info.get_attribute_uint64("time::modified")
     except GLib.Error as e:
         if e.code != Gio.IOErrorEnum.NOT_FOUND:
-            print("PearlInstall: flatpak - could not get time::modified from file %s" % gfile.get_path())
+            print("Installer: flatpak - could not get time::modified from file %s" % gfile.get_path())
         return 0
 
 def _should_update_appstream_data(fp_sys, remote, arch):
@@ -90,14 +117,14 @@ def _should_update_appstream_data(fp_sys, remote, arch):
 
     try:
         if fp_sys.update_remote_sync(remote.get_name(), None):
-            print("PearlInstall: flatpak - metadata for remote '%s' has been updated. Comparing appstream timestamps..." % remote.get_name())
+            print("Installer: flatpak - metadata for remote '%s' has been updated. Comparing appstream timestamps..." % remote.get_name())
 
             new_timestamp = _get_file_timestamp(remote.get_appstream_timestamp(arch))
 
             if (new_timestamp > current_timestamp) or (current_timestamp == 0):
                 ret = True
     except GLib.Error as e:
-        print("PearlInstall: flatpak - could not update metadata for remote '%s': %s" % (remote.get_name(), e.message))
+        print("Installer: flatpak - could not update metadata for remote '%s': %s" % (remote.get_name(), e.message))
 
     return ret
 
@@ -105,13 +132,13 @@ def _process_remote(cache, fp_sys, remote, arch):
     remote_name = remote.get_name()
 
     if remote.get_disabled():
-        print("PearlInstall: flatpak - remote '%s' is disabled, skipping" % remote_name)
+        print("Installer: flatpak - remote '%s' is disabled, skipping" % remote_name)
         return
 
     remote_url = remote.get_url()
 
     if _should_update_appstream_data(fp_sys, remote, arch):
-        print("PearlInstall: flatpak - new appstream data available for remote '%s', fetching..." % remote_name)
+        print("Installer: flatpak - new appstream data available for remote '%s', fetching..." % remote_name)
 
         try:
             fp_sys.update_appstream_sync(remote_name, arch, None)
@@ -119,12 +146,12 @@ def _process_remote(cache, fp_sys, remote, arch):
             # Not fatal..
             pass
     else:
-        print("PearlInstall: flatpak - no new appstream data for remote '%s', skipping download" % remote_name)
+        print("Installer: flatpak - no new appstream data for remote '%s', skipping download" % remote_name)
 
     # get_noenumerate indicates whether a remote should be used to list applications.
     # Instead, they're intended for single downloads (via .flatpakref files)
     if remote.get_noenumerate():
-        print("PearlInstall: flatpak - remote '%s' is marked as no-enumerate skipping package listing" % remote_name)
+        print("Installer: flatpak - remote '%s' is marked as no-enumerate skipping package listing" % remote_name)
         return
 
     try:
@@ -135,6 +162,9 @@ def _process_remote(cache, fp_sys, remote, arch):
             if ref.get_arch() != arch:
                 continue
 
+            if ref.get_name().endswith("BaseApp"):
+                continue
+
             _add_package_to_cache(cache, ref, remote_url, False)
     except GLib.Error as e:
         print(e.message)
@@ -184,10 +214,10 @@ def process_full_flatpak_installation(cache):
             flatpak_remote_infos[remote_name] = FlatpakRemoteInfo(remote)
 
     except GLib.Error as e:
-        print("PearlInstall: flatpak - could not get remote list", e.message)
+        print("Installer: flatpak - could not get remote list", e.message)
         cache = {}
 
-    print('PearlInstall: Processing Flatpaks for cache took %0.3f ms' % ((time.time() - fp_time) * 1000.0))
+    print('Installer: Processing Flatpaks for cache took %0.3f ms' % ((time.time() - fp_time) * 1000.0))
 
     return cache, flatpak_remote_infos
 
@@ -215,7 +245,7 @@ def _initialize_appstream_thread():
             for remote in fp_sys.list_remotes():
                 _load_appstream_pool(_as_pools, remote)
         except GLib.Error:
-            print("PearlInstall: Could not initialize appstream components for flatpaks")
+            print("Installer: Could not initialize appstream components for flatpaks")
 
 def search_for_pkginfo_as_component(pkginfo):
     name = pkginfo.name
@@ -452,7 +482,12 @@ def _get_remote_related_refs(fp_sys, remote_name, ref):
 
     return return_refs
 
-def _get_theme_refs(fp_sys, remote_name, ref):
+def _get_theme_refs(fp_sys, remote_name, ref=None):
+    if ref:
+        arch = ref.get_arch()
+    else:
+        arch = Flatpak.get_default_arch()
+
     theme_refs = []
 
     gtksettings = Gtk.Settings.get_default()
@@ -489,7 +524,7 @@ def _get_theme_refs(fp_sys, remote_name, ref):
             matching_refs = sorted(matching_refs, key=sortref, reverse=True)
 
             for matching_ref in matching_refs:
-                if matching_ref.get_arch() != ref.get_arch():
+                if matching_ref.get_arch() != arch:
                     continue
 
                 theme_ref = matching_ref
@@ -521,7 +556,7 @@ def _get_theme_refs(fp_sys, remote_name, ref):
                     matching_refs = sorted(matching_refs, key=sortref, reverse=True)
 
                     for matching_ref in matching_refs:
-                        if matching_ref.get_arch() != ref.get_arch():
+                        if matching_ref.get_arch() != arch:
                             continue
 
                         theme_ref = matching_ref
@@ -541,6 +576,20 @@ def _get_theme_refs(fp_sys, remote_name, ref):
 
     return theme_refs
 
+def get_updated_theme_refs():
+    fp_sys = get_fp_sys()
+
+    if not fp_sys.list_installed_refs_by_kind(Flatpak.RefKind.APP, None):
+        return []
+
+    theme_refs = []
+
+    for remote in fp_sys.list_remotes():
+        theme_refs += _get_theme_refs (fp_sys, remote.get_name())
+
+    return theme_refs
+
+
 def _get_installed_related_refs(fp_sys, remote, ref):
     return_refs = []
 
@@ -594,7 +643,7 @@ def select_packages(task):
     else:
         method = _pick_refs_for_removal
 
-    print("PearlInstall: Calculating changes required for Flatpak package: %s" % task.pkginfo.name)
+    print("Installer: Calculating changes required for Flatpak package: %s" % task.pkginfo.name)
 
     thread = threading.Thread(target=method, args=(task,))
     thread.start()
@@ -700,7 +749,7 @@ def list_updated_pkginfos(cache):
     try:
         updates = fp_sys.list_installed_refs_for_update(None)
     except GLib.Error as e:
-        print("PearlInstall: flatpak - could not get updated flatpak refs")
+        print("Installer: flatpak - could not get updated flatpak refs")
         return []
 
     for ref in updates:
@@ -736,7 +785,7 @@ def generate_uncached_pkginfos(cache):
                     _add_package_to_cache(cache, ref, remote.get_url(), True)
 
     except GLib.Error as e:
-        print("PearlInstall: flatpak - could not check for uncached pkginfos", e.message)
+        print("Installer: flatpak - could not check for uncached pkginfos", e.message)
 
 def pkginfo_is_installed(pkginfo):
     fp_sys = get_fp_sys()
@@ -765,7 +814,7 @@ def list_remotes():
             remotes.append(FlatpakRemoteInfo(remote))
 
     except GLib.Error as e:
-        print("PearlInstall: flatpak - could not fetch remote list", e.message)
+        print("Installer: flatpak - could not fetch remote list", e.message)
         remotes = []
 
     return remotes
@@ -780,11 +829,12 @@ def _pkginfo_from_file_thread(cache, file, callback):
     path = file.get_path()
 
     if path == None:
-        print("PearlInstall: flatpak - no valid .flatpakref path provided")
+        print("Installer: flatpak - no valid .flatpakref path provided")
         return None
 
     ref = None
     pkginfo = None
+    remote_name = None
 
     with open(path) as f:
         contents = f.read()
@@ -793,33 +843,45 @@ def _pkginfo_from_file_thread(cache, file, callback):
         gb = GLib.Bytes(b)
 
         try:
-            ref = fp_sys.install_ref_file(gb, None)
-        except GLib.Error as e:
-            if e.code == Flatpak.Error.ALREADY_INSTALLED:
-                try:
-                    kf = GLib.KeyFile()
-                    if kf.load_from_file(path, GLib.KeyFileFlags.NONE):
-                        name = kf.get_string("Flatpak Ref", "Name")
-                        branch = kf.get_string("Flatpak Ref", "Branch")
-                        url = kf.get_string("Flatpak Ref", "Url")
-                        if name and branch and url:
-                            basic_ref = Flatpak.Ref.parse("app/%s/%s/%s" % (name, Flatpak.get_default_arch(), branch))
-                            remote_name = _get_remote_name_by_url(fp_sys, url)
+            kf = GLib.KeyFile()
+            if kf.load_from_file(path, GLib.KeyFileFlags.NONE):
+                name = kf.get_string("Flatpak Ref", "Name")
+                url = kf.get_string("Flatpak Ref", "Url")
 
-                            ref = Flatpak.RemoteRef(remote_name=remote_name,
-                                                    kind=basic_ref.get_kind(),
-                                                    arch=basic_ref.get_arch(),
-                                                    branch=basic_ref.get_branch(),
-                                                    name=basic_ref.get_name())
-                except GLib.Error:
-                    print("PearlInstall: flatpak package already installed, but an error occurred finding it")
-            elif e.code != Gio.DBusError.ACCESS_DENIED: # user cancelling auth prompt for adding a remote
-                print("PearlInstall: could not read .flatpakref file: %s" % e.message)
-                dialogs.show_flatpak_error(e.message)
+                try:
+                    branch = kf.get_string("Flatpak Ref", "Branch")
+                except GLib.Error as e:
+                    if e.code == GLib.KeyFileError.KEY_NOT_FOUND:
+                        print("Installer: flatpak - flatpakref file doesn't have a Branch key, maybe nightly or testing.")
+                        branch = None
+
+                remote_name = _get_remote_name_by_url(fp_sys, url)
+
+                if name and remote_name:
+                    ref = Flatpak.RemoteRef(remote_name=remote_name,
+                                            kind=Flatpak.RefKind.APP,
+                                            arch=Flatpak.get_default_arch(),
+                                            branch=branch,
+                                            name=name)
+                    print("Installer: flatpak - using existing remote '%s' for flatpakref file install" % remote_name)
+                else: #If Flatpakref is not installed already
+                    try:
+                        print("Installer: flatpak - trying to install new remote for flatpakref file")
+                        ref = fp_sys.install_ref_file(gb, None)
+                        fp_sys.drop_caches(None)
+
+                        remote_name = ref.get_remote_name()
+                        print("Installer: flatpak - added remote '%s'" % remote_name)
+                    except GLib.Error as e:
+                        if e.code != Gio.DBusError.ACCESS_DENIED: # user cancelling auth prompt for adding a remote
+                            print("Installer: could not add new remote to system: %s" % e.message)
+                            dialogs.show_flatpak_error(e.message)
+        except GLib.Error as e:
+            print("Installer: flatpak - could not parse flatpakref file: %s" % e.message)
+            dialogs.show_flatpak_error(e.message)
 
         if ref:
             try:
-                remote_name = ref.get_remote_name()
                 remote = fp_sys.get_remote_by_name(remote_name, None)
 
                 # Add the ref's remote if it doesn't already exist
@@ -868,13 +930,12 @@ def _pkginfo_from_file_thread(cache, file, callback):
                             for remote in fp_sys.list_remotes(None):
                                 # See comments below in _remote_from_repo_file_thread about get_noenumerate() use.
                                 if remote.get_url() == runtime_repo_url and not remote.get_noenumerate():
-                                    print("PearlInstall: runtime remote '%s' already in system, skipping" % runtime_remote_name)
+                                    print("Installer: flatpak - runtime remote '%s' already in system, skipping" % runtime_remote_name)
                                     existing = True
                                     break
 
-                            # if not existing:
-                            if True:
-                                print("PearlInstall: Adding additional runtime remote named '%s' at '%s'" % (runtime_remote_name, runtime_repo_url))
+                            if not existing:
+                                print("Installer: Adding additional runtime remote named '%s' at '%s'" % (runtime_remote_name, runtime_repo_url))
 
                                 cmd_v = ['flatpak',
                                          'remote-add',
@@ -888,7 +949,7 @@ def _pkginfo_from_file_thread(cache, file, callback):
                                 fp_sys.drop_caches(None)
                         os.unlink(file.name)
             except GLib.Error as e:
-                print("PearlInstall: could not process .flatpakref file: %s" % e.message)
+                print("Installer: could not process .flatpakref file: %s" % e.message)
                 dialogs.show_flatpak_error(e.message)
 
     GLib.idle_add(callback, pkginfo, priority=GLib.PRIORITY_DEFAULT)
@@ -901,7 +962,7 @@ def _remote_from_repo_file_thread(cache, file, callback):
     try:
         path = Path(file.get_path())
     except TypeError:
-        print("PearlInstall: flatpak - no valid .flatpakrepo path provided")
+        print("Installer: flatpak - no valid .flatpakrepo path provided")
         return
 
     fp_sys = get_fp_sys()
@@ -1001,12 +1062,12 @@ class MetaTransaction():
         self.item_count = len(task.to_install + task.to_remove + task.to_update)
         self.current_count = 0
 
-        try:
-            for ref in task.to_install:
-                task.progress_state = task.PROGRESS_STATE_INSTALLING
-                task.current_package_name = ref.get_name()
+        for ref in task.to_install:
+            task.progress_state = task.PROGRESS_STATE_INSTALLING
+            task.current_package_name = ref.get_name()
 
-                print("installing: %s" % ref.format_ref())
+            print("installing: %s" % ref.format_ref())
+            try:
                 self.fp_sys.install(ref.get_remote_name(),
                                     ref.get_kind(),
                                     ref.get_name(),
@@ -1015,14 +1076,24 @@ class MetaTransaction():
                                     self.on_flatpak_progress,
                                     None,
                                     task.cancellable)
+            except GLib.Error as e:
+                if task.current_package_name == task.pkginfo.name:
+                    if e.code != Gio.IOErrorEnum.CANCELLED:
+                        task.progress_state = task.PROGRESS_STATE_FAILED
+                        task.current_package_name = None
+                        self.on_flatpak_error(e.message)
+                        return
+                else:
+                    print("Installer: flatpak - problem installing %s: %s" % (task.current_package_name, e.message))
 
-                self.current_count += 1
+            self.current_count += 1
 
-            for ref in task.to_remove:
-                task.progress_state = task.PROGRESS_STATE_REMOVING
-                task.current_package_name = ref.get_name()
+        for ref in task.to_remove:
+            task.progress_state = task.PROGRESS_STATE_REMOVING
+            task.current_package_name = ref.get_name()
 
-                print("removing: %s" % ref.format_ref())
+            print("removing: %s" % ref.format_ref())
+            try:
                 self.fp_sys.uninstall(ref.get_kind(),
                                       ref.get_name(),
                                       ref.get_arch(),
@@ -1030,14 +1101,24 @@ class MetaTransaction():
                                       self.on_flatpak_progress,
                                       None,
                                       task.cancellable)
+            except GLib.Error as e:
+                if task.current_package_name == task.pkginfo.name:
+                    if e.code != Gio.IOErrorEnum.CANCELLED:
+                        task.progress_state = task.PROGRESS_STATE_FAILED
+                        task.current_package_name = None
+                        self.on_flatpak_error(e.message)
+                        return
+                else:
+                    print("Installer: flatpak - problem removing %s: %s" % (task.current_package_name, e.message))
 
-                self.current_count += 1
+            self.current_count += 1
 
-            for ref in task.to_update:
-                task.progress_state = task.PROGRESS_STATE_UPDATING
-                task.current_package_name = ref.get_name()
+        for ref in task.to_update:
+            task.progress_state = task.PROGRESS_STATE_UPDATING
+            task.current_package_name = ref.get_name()
 
-                print("updating: %s" % ref.format_ref())
+            print("updating: %s" % ref.format_ref())
+            try:
                 self.fp_sys.update(Flatpak.UpdateFlags.NONE,
                                    ref.get_remote_name(),
                                    ref.get_kind(),
@@ -1047,14 +1128,11 @@ class MetaTransaction():
                                    self.on_flatpak_progress,
                                    None,
                                    task.cancellable)
+            except GLib.Error as e:
+                # update will None as a task's pkginfo, just warn.
+                print("Installer: flatpak - problem updating %s: %s" % (task.current_package_name, e.message))
 
-                self.current_count += 1
-        except GLib.Error as e:
-            if e.code != Gio.IOErrorEnum.CANCELLED:
-                task.progress_state = task.PROGRESS_STATE_FAILED
-                task.current_package_name = None
-                self.on_flatpak_error(e.message)
-                return
+            self.current_count += 1
 
         task.progress_state = task.PROGRESS_STATE_FINISHED
         task.current_package_name = None
